def activity_to_plain_text(activity : dict) -> str:
    """Convert activity dictionary to plain

    Args:
        activity (dict): activity dict from scrapper (single one)

    Returns:
        str: plain
    """
    return f"{activity['title']}\n\tGoal: {activity['goal']}\n\tResults: {activity['results']}\n\tIssue_encountered: {activity['issue_encountered']}\n\tSkills mastered: {activity['skills']}\n\tFacts: {activity['facts']}\n"



def activities_dict_to_items(act_dict: dict) -> list:
    """Convert activity dictionary to plain list of activity items (dict)

    Args:
        act_dict (dict): activity dict from scrapper

    Returns:
        list: list of all activity
    """
    single_activity_list = []
    for activity_list in act_dict.values():
        single_activity_list.extend(activity_list)
    return single_activity_list

def activities_dict_to_plain_text(act_dict: dict) -> str:
    """Convert activity dictionary to plain text

    Args:
        act_dict (dict): activity dict from scrapper

    Returns:
        str: Plain text of all activity contained in dict
    """
    single_text = ""
    for activity in activities_dict_to_items(act_dict):
        single_text += activity_to_plain_text(activity)
    return single_text


import asyncio
import datetime
import aiohttp
import bs4
import re
from tqdm.asyncio import tqdm

COOKIE_KEY = "PHPSESSID"
LINK_CLASS = "item"
LINK_FILTER = "afficher/periode"
ACTIVITY_LIST_CLASS = "ui tab scrollable-tab"
REPORT_FILTER = "ACTIVIT"
DESCRIPTION_CLASS = "description"
BOOK_CLASS = "ui no-disabled-opacity form"
END_OF_ACTIVITY = "Pièces jointes du questionnaire"
ACTIVITIES_DIV_CLASS = "field"
ACTIVITY_DIV_CLASS = "accordion"
CONTENT_ACTIVE_CLASS = "content active"
FIELD_REPLACE = "[Moyenne des réponses : ]"
class LoginError(Exception):
    pass

class ParsingError(Exception):
    pass

class Scraper:
    def __init__(self, 
                 url: str,
                 email: str = None, 
                 password: str = None, 
                 cookie: str = None, 
                 verbose: bool = False):
        if not url.startswith('http://') or not url.startswith('https://'):
            url = f"https://{url}"
        self.url = url
        self.email = email
        self.password = password
        self.cookie = cookie
        self.verbose = verbose
        self.session = None
    async def __aenter__(self):
        """Async context manager."""
        # Create a new session when entering the context manager.
        self.session = aiohttp.ClientSession()
        try:
            await self.login()
        except Exception as e:
            await self.session.close()
            raise e
        return self
    
    async def __aexit__(self, exc_type, exc, tb):
        """Async context manager."""
        # Close the session when exiting the context manager.
        await self.session.close()
    
    async def login(self):

        if self.cookie:
            """Cookie authentication."""
            cookie_dict = {COOKIE_KEY: self.cookie}
            self.session.cookie_jar.update_cookies(cookie_dict, self.url)
            if self.verbose:
                print("Testing the cookie...")
            async with self.fetch(f"{self.url}/_connexion") as response:
                if "Connexion" in response:
                    raise LoginError("The cookie provided is invalid")
                
        elif self.email:
            """Email authentication."""
            if self.verbose:
                print("Logging in with email and password...")
            data={
                "_username": self.email,
                "_password": self.password,
                "_remember_me": 1,
            }
            login_response = await self.session.post(
                f"{self.url}/_connexion",
                data=data
            )
            if login_response.status != 200:
                if self.verbose:
                    print(f"Login request failed with status code {login_response.status}")
                raise LoginError(f"The login request failed with status code {login_response.status}")
            elif "Identifiants invalides" in await login_response.text():
                raise LoginError("Invalid email or password.")
            else:
                if self.verbose:
                    print("Login successful.")
            
        else:
            raise ValueError("Please provide either a cookie or an email and password.")

    async def fetch(self, full_url: str = None):
        if self.verbose:
            print(f"Fetching {full_url}...")
        async with self.session.get(full_url) as response:
            return await response.text(encoding="utf-8")
    
    
    async def list_activities(self):
        """List the activities."""
        if self.verbose:
            print("Listing the activities...")
        response = await self.fetch(f"{self.url}/book/liste#/alll")
        soup = bs4.BeautifulSoup(response, "html.parser")
        all_pages_list = soup.find_all("div", class_=ACTIVITY_LIST_CLASS)[-1]

        all_pages_links = all_pages_list.find_all("a", class_=LINK_CLASS)

        filtered_links = {}
        for link in all_pages_links:
            if REPORT_FILTER in link.text and LINK_FILTER in link["href"]:
                description = link.find('div', class_=DESCRIPTION_CLASS)
                description = description.text.strip()
                date = description[-10:]
                try:
                    date_obj = datetime.datetime.strptime(date, "%d/%m/%Y")
                except ValueError as e:
                    raise ParsingError(f"Error while parsing the date: {e}")
                filtered_links[date_obj] = link["href"]
                if self.verbose:
                    print(f"Found activity due to {date} ({self.url}{link['href']})")
        return filtered_links
    
    def _clean_text(self, text: str) -> str:
        text = text.replace("\n ", "\n")
        text = text.strip()
        
       
        # remove tabs
        text = re.sub(r"\t+", "", text)
        # remove multiple spaces
        text = re.sub(r" +", " ", text)

        # remove multiple newlines
        text = re.sub(r"\n+", "\n", text)
        
        # interprete all unicode escape sequences
        #  \u00e0 transform to à
        # replace all \
        # use a lambda function to replace all unicode escape sequences
        
        
        
        return text
    
    def _clean_field(self, field: bs4.element.Tag) -> str:
        # decode
        text = field.text.strip()
        text = text.replace(FIELD_REPLACE, "")
        text = self._clean_text(text)
        return text
    
    def _parse_activity(self, activity_div: bs4.element.Tag) -> dict:
        activity = {}
        activity_fields = []
        new_activity = False
        try:
            activity_fields = activity_div.find('div', class_=CONTENT_ACTIVE_CLASS).find_all('div', class_=CONTENT_ACTIVE_CLASS)
        except AttributeError as e:
            activity_fields = activity_div.find_all('div', class_="ui segment")
            new_activity = True
        cleaned_fields = []
        i = 0
        for field in activity_fields:
            if not new_activity:
                if i >= 2:
                    field = field.find_all('div', class_="field")[1].find('div', class_="ui segment")
            i+=1
            cleaned_fields.append(self._clean_field(field))
        if new_activity:
            activity["title"] = cleaned_fields[0]
            activity["goal"] = cleaned_fields[1]
            activity["results"] = cleaned_fields[2]
            activity["skills"] = cleaned_fields[3]
            activity["issue_encountered"] = cleaned_fields[4]
            activity["facts"] = cleaned_fields[5]
        else:
            activity["title"] = f"{cleaned_fields[0]} ({cleaned_fields[1]})"
            activity["goal"] = cleaned_fields[2]
            activity["results"] = cleaned_fields[3]
            activity["skills"] = cleaned_fields[4]
            activity["issue_encountered"] = cleaned_fields[5]
            activity["facts"] = cleaned_fields[6]
        return activity

    def _parse_report(self, report_div: bs4.element.Tag) -> dict:
        activities_divs = report_div.find("div", class_=ACTIVITIES_DIV_CLASS).find("div")
        individual_activities = activities_divs.find_all("div", recursive=False)
        if len(individual_activities) == 0:
            activities_divs = report_div.find_all("div", class_=ACTIVITIES_DIV_CLASS)[1].find("div").find("div").find("div")
            individual_activities = activities_divs.find_all("div", class_="ui segment field")
        activities = []
        for activity_div in activities_divs.find_all("div", recursive=False):
            activities.append(self._parse_activity(activity_div))
        return activities

    async def get_all_activity(self):
        """Get all the activities."""
        activities = await self.list_activities()
        futures = [self.fetch(f"{self.url}{link}") for link in activities.values()]
        responses = []
        if self.verbose:
            responses = await tqdm.gather(*futures)
        else:
            responses = await asyncio.gather(*futures)
        act_text = {}
        for i in range(len(responses)):
            title = list(activities.keys())[i]
            if self.verbose:
                print(f"Parsing activity: {title}")
            response = responses[i]
            soup = bs4.BeautifulSoup(response, "html.parser")
            book = soup.find("form", class_=BOOK_CLASS)
            act_text[title] = self._parse_report(book)
           
        
        # sort the activities by date
        act_text = dict(sorted(act_text.items()))
        return act_text



""" Command line interface for the calisum package. """

import argparse
import json
import bullet
from getpass import getpass

import tabulate
from calisum import PARSING_ERROR, SCRAPING_ERROR, __app_name__, __version__
from calisum.scraper import LoginError, ParsingError, Scraper, COOKIE_KEY
from calisum.summarizer import Summarizer, Models
from calisum.util import activities_dict_to_plain_text, activities_dict_to_items

KEYS_SHOWS_LLM_PICK_MENU = [
    "id",
    "name",
    "version"
]

def show_version():
    """Print the version of the package."""
    print(f"{__app_name__} {__version__}")

def setup_parser():
    """Create the command line interface parser."""
    parser = argparse.ArgumentParser(
        prog=__app_name__,
        description="Summarize Caliap activity data."
    )
    parser.add_argument(
        "--version",
        action="store_true",
        help="Display the version of the package."
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Increase the verbosity of the output."
    )
    parser.add_argument(
        "-O",
        "--output",
        help="Output file to save the summary to. default to stdout "
    )
    parser.add_argument(
        "-o",
        "--original-output",
        action="store_true",
        help="Output the original data instead of the summary.")
    parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        help='Output the result as json'
    )

    scraper = parser.add_argument_group("Scraper options")

    scraper.add_argument(
        "url_to_scrape",
        help="URL to scrape the data from.",
        nargs="?"
    )
    auth_group = scraper.add_mutually_exclusive_group()

    auth_group.add_argument(
        "-e",
        "--email",
        help="Email address to use for authentication. Must be used together with --password."
    )
    scraper.add_argument(
        "-p",
        "--password",
        help="Password to use for authentication. Must be used together with --email."
    )
    auth_group.add_argument(
        "-c",
        "--cookie",
        help=f"Cookie to use for authentication ({COOKIE_KEY}). Mutually exclusive with --email and --password."
    )
    llm = parser.add_argument_group("LLM options")
    llm.add_argument(
        "-l",
        "--llm-url",
        help="URL to the LLM page open api endpoint (compatible with openai api standard)."
    )
    llm.add_argument(
        "-t",
        "--llm-token",
        help="LLM token (usefull with openai)"
    )
    llm.add_argument(
        "-n",
        "--jan-ai",
        action="store_true",
        help="Use localhost default jan-ai config."
    )
    llm.add_argument(
        "-m",
        "--model-id",
        help="OpenAI form model id (eg: mistral-ins-7b-q4)"
    )
    llm.add_argument(
        "-g",
        "--global-summary",
        help="Do a global summary of individual activity (works only if original output is not selected)",
        action="store_true"
    )
    return parser
    
def dict_to_menu(input_dicts : list[dict[str, str]]) -> dict[str,str]:
    """Shows a menu in stdout to pick one value of the dict

    Args:
        input_dict (list): list of dict 

    Returns:
        dict: Selected dictionnary
    """
    table = tabulate.tabulate(input_dicts, headers='keys')
    table_lines = table.split('\n')
    prompt = str.join("\n", table_lines[:2])
    picker = bullet.Bullet(choices=table_lines[2:], bullet="•", prompt=str.join("\n", table_lines[:2]), return_index=True)
    print(prompt)
    choice = picker.launch()[1] # Grab index from returned tuple
    return input_dicts[choice]
    
    
def strip_values(input_dicts : dict) -> dict:
    new_dict = {}
    for key in input_dicts:
        if key in KEYS_SHOWS_LLM_PICK_MENU:
            new_dict[key] = input_dicts[key]
    return new_dict


async def main():
    """Run the command line interface."""
    parser = setup_parser()
    args = parser.parse_args()
    if args.version:
        show_version()
        return
    elif args.url_to_scrape is None:
        print("Please provide a URL to scrape.")
        print("Use --help for more information.")
        return 
    elif args.jan_ai and args.llm_url:
        print("You cannot use both --jan-ai and --llm_url")
        exit(PARSING_ERROR)
    if args.email:
        """Email authentication."""
        password = args.password
        if password is None:
            password = getpass("Enter your password for {}: ".format(args.email))
    else:
        """Cookie authentication."""
        if args.cookie is None:
            print("Please provide a cookie for authentication or use email and password.")
            exit(PARSING_ERROR)
    if args.original_output and args.global_summary:
        print("--original-output and --global-summary are mutually exclusive.")
        exit(PARSING_ERROR)
    try:
        async with Scraper(
            url=args.url_to_scrape,
            email=args.email,
            password=password,
            cookie=args.cookie,
            verbose=args.verbose
        ) as scraper:
            if args.verbose:
                print("Scraping the data...")
            activities_dict = await scraper.get_all_activity()
            if args.original_output:
                # No summary
                text = activities_dict_to_plain_text(activities_dict)
                if args.json:
                    text = json.dumps(
                        activities_dict_to_items(activities_dict),
                        indent=3
                    )
                if args.output:
                    with open(args.output, "w") as file:
                        file.write(text)
                else:
                    print(text)
                
            else:
                # Part where we summarize the data
                model = Models.SUMY
                model_url = ""
                model_token = ""
                text = ""
                if args.llm_url is not None and args.jan_ai is False:
                    model = Models.CUSTOM
                    model_url = args.llm_url
                    model_token = args.llm_token
                elif args.jan_ai is True and args.llm_url is None:
                    model = Models.JANAI
                elif args.llm_url is not None:
                    model= Models.CUSTOM
                async with Summarizer(
                    verbose=args.verbose,
                    model=model,
                    model_url=model_url,
                    model_token=model_token,
                ) as summarizer:
                    if model in [Models.JANAI, Models.CUSTOM, Models.CHATGPT]:
                        # Ask using curse
                        if args.model_id is None:
                            llm_list = [strip_values(model) for model in await summarizer.list_llm_models()]
                            llm_selected = dict_to_menu(llm_list)
                            await summarizer.set_llm_id(llm_selected['id'])
                        else:
                            await summarizer.set_llm_id(args.model_id)
                    texts = await summarizer.summarize_activities(activities_dict_to_items(activities_dict))
                    if args.global_summary:
                        if args.verbose:
                            print("Summarizing globally")
                        await summarizer.summarize_global(texts)
                    if args.json:
                        text = json.dumps(texts, indent=3)
                    else:
                        text = "\n".join(texts)
                if args.output:
                    with open(args.output, "w") as file:
                        file.write(text)
                else:
                    print(text)
    except LoginError as e:
        if args.verbose:
            print(f"Error while logging in: {e}")
        exit(SCRAPING_ERROR)
    except ParsingError as e:
        if args.verbose:
            print(f"Error while parsing the data: {e}")
        exit(PARSING_ERROR)


"""Caliap Activity Summarizer package."""

__app_name__ = "calisum"
__version__ = "1.1.1"

(
    SUCCESS,
    ERROR,
    SCRAPING_ERROR,
    PARSING_ERROR,
    SUMMARIZING_ERROR,
) = range(5)

ERROR_MESSAGES = {
    SUCCESS: "Success",
    ERROR: "Error",
    SCRAPING_ERROR: "Error while scraping",
    PARSING_ERROR: "Error while parsing",
    SUMMARIZING_ERROR: "Error while summarizing",
}


import asyncio
import aiohttp
import nltk
from sumy.parsers.plaintext import PlaintextParser
from sumy.nlp.tokenizers import Tokenizer
from sumy.summarizers.lsa import LsaSummarizer
from openai import AsyncOpenAI
from tqdm.asyncio import tqdm
from calisum.summarizer.models import Models
from calisum.util import activity_to_plain_text

DEFAULT_LANG = 'fr'
DEFAULT_JANAI_API_ENDPOINT = 'http://localhost:1337/v1'
SUMMY_SENTENCE_OUTPUT = 2 # Number of sentences to return per activity
DUMMY_TOKEN = "NOTATOKEN"
INDIVIDIUAL_PROMPT = "Résume cette activité ci-dessous en quelques phrases essentielles : En quoi consiste grossièrement la tache et qu'est ce qu'elle m'a apportée ?"
SEPARATOR = '\n___\n'
GLOBAL_PROMPT= f"Fait un résumé cohérent de l'ensemble de ces activités (celles si son séparées par '{SEPARATOR}' et classé par ordre chronologique). Déduit une évolution et prend du recul par rapport à celles-ci"


class SummarizerException(Exception):
    pass

class SummarizerNotConfiguredException(SummarizerException):
    pass

class Summarizer:
    def __init__(
            self,
            verbose: bool = False, 
            model: int = Models.SUMY, 
            model_url : str = "",
            model_token: str = DUMMY_TOKEN,
        ):
        """Init the summarizer with the chosen model and settings

        Args:
            verbose (bool, optional): If set to True will log what summarizer is doing. Defaults to False.
            model (int, optional): Choose a model from Models enum. Defaults to Models.SUMY.
            model_url (str, optional): Url of the model api to request. Defaults to "".
            model_token (str, optional): Token of the model api to request. Defaults to "".
        """
        self.verbose = verbose
        if model == Models.JANAI and model_url == "":
            model_url = DEFAULT_JANAI_API_ENDPOINT
        self.url = model_url
        self.token = model_token if model_token is not None else DUMMY_TOKEN
        self.__llm_id = None

        self.model = model
        if model != Models.SUMY:
            self.llm = AsyncOpenAI(
                api_key=self.token,
                base_url=self.url
            )
        self._func_dict = {
            Models.SUMY : self.__sum_sumy,
            Models.CHATGPT: self.__sum_llm,
            Models.JANAI: self.__sum_llm,
            Models.CUSTOM: self.__sum_llm
        }
    @property
    def llm_id(self):
        return self.__llm_id

    async def set_llm_id(self, value):
        # Checks that id is one of the value possible
        list_of_id = []
        for model in await self.list_llm_models():
            list_of_id.append(model['id'])
        
        if value not in list_of_id:
            raise ValueError(f"{value} not in available model id")
        self.__llm_id = value


    async def __aenter__(self):
        """context manager."""
        # Create a new session when entering the context manager.
        if self.model == Models.SUMY:
            # Install tokenizer
            nltk.download('punkt_tab')
            self.tokenizer = Tokenizer(DEFAULT_LANG)
            self.summarizer = LsaSummarizer()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        """context manager."""
        pass

    async def summarize_activities(self, activities : list[dict]) -> list[str]:
        """Summarize activities individualy 

        Args:
            activities (list[dict]): List of activities to summarize

        Returns:
            list[str]: Return a list of summarized activity
        """ 
        futures = []
        output = []
        for activity in activities:
            text = activity_to_plain_text(activity)
            futures.append(self._func_dict[self.model](text))
        if self.verbose:
            output = await tqdm.gather(*futures)
        else:
            output = await asyncio.gather(*futures)
        return output
    
    async def list_llm_models(self) -> list[dict]:
        """List available llm models on the remote server.
        Models should be one of JAN, CUSTOM, or CHATGPT

        Returns:
            list[dict]: List of available llm models on the dictionary
        """
        if self.model not in [Models.CHATGPT, Models.CUSTOM, Models.JANAI]:
            raise SummarizerException(f"Model {self.model} does not support listing model")
        models_obj = list(await self.llm.models.list())[0][1]
        output = [dict(model_obj) for model_obj in models_obj]
        return output
    
    async def summarize(self, text : str) -> str :
        """Summarize the text

        Args:
            text (str, optional): Text to summarize.

        Returns:
            str: Summary of the text
        """
        return self._func_dict[self.model](text)

    async def __sum_sumy(self, text : str, nb_of_sentence: int = SUMMY_SENTENCE_OUTPUT) -> str:
        """Summarize with summy

        Args:
            text (str): Text to summarize

        Returns:
            str: Summary
        """
        parser = PlaintextParser.from_string(text, self.tokenizer)
        summarizer = self.summarizer(parser.document, nb_of_sentence)
        return''.join([str(sentence) for sentence in summarizer])

    async def __sum_llm(self, text : str, prompt: str = INDIVIDIUAL_PROMPT) -> str:
        """Summarize using llm requires a model id
        will fail if llm_id is not defined

        Args:
            text (str): Text to summarize

        Returns:
            str: Summarized text
        """
        if self.llm_id is None :
            raise SummarizerNotConfiguredException('llm_id is not defined')

        payload = {
            'role' : 'user',
            'content' : prompt + ' ' + text
        }
        response = await self.llm.chat.completions.create(messages=[payload], model=self.llm_id)
        return response.to_dict()['choices'][0]['message']['content']

    async def summarize_global(self, text : list[str]) -> str:
        """Summarize a list of texts globally

        Args:
            text (list[str]): List of texts to summarize

        Returns:
            str: Summarized text
        """
        merged_text = SEPARATOR.join(text)
        if self.model == Models.SUMY:
            return await self.__sum_sumy(merged_text, len(text)//5 + 1)
        else :
            return await self.__sum_llm(merged_text, GLOBAL_PROMPT)